from enbios2.const import BASE_DATA_PATH

scenario_col = "Scenario"
value_col = "Value"
level_col = "Dendrogram level"
processor_col = "Processor"
indicator_col = "Indicator"

ABSOLUTE = "absolute"
RELATIVE = "relative"


INDICATOR = "indicator"
PROCESSOR = "processor"
SCENARIO = "scenario"

# the filee which includes the information about the indicators
indicator_map_file = BASE_DATA_PATH / "generic/indicator_info.json"
